"""
The ratelimiter used by the metrics string indexer to rate-limit DB writes.

As opposed to the API rate limiter, a limit in the "sliding window" rate
limiter such as "10 requests / minute" does not reset to 0 every minute.
Instead each window can be configured with a "granularity" setting such that
the window gradually resets in steps of `granularity` seconds.

Additionally this rate-limiter is not coupled to per-project/organization
scopes, and can apply multiple sliding windows at once. On the flipside it is
not strongly consistent and depending on usage it is very easy to over-spend
quota, as checking quota and spending quota are two separate steps.

Example
=======

We want to enforce the number of requests per organization via two limits:

* 100 per 30-second
* 10 per 3-seconds

On every request, our API endpoint calls::

    check_and_use_quotas(RequestedQuota(
        prefix=f"org-id:{org_id}"
        quotas=[
            Quota(
                window_seconds=30,
                limit=100,

                # can be arbitrary depending on how "sliding" the sliding
                # window should be. This one configures per-10-second granularity
                # to make the example simpler
                granularity=10,
            ),
            Quota(
                window_seconds=3,
                limit=10,
                granularity=1,
            )
        ]
    ))

For a request happening at time `900` for `org_id=123`, the redis backend
checks the following keys::

    sliding-window-rate-limit:123:3:900
    sliding-window-rate-limit:123:3:899
    sliding-window-rate-limit:123:3:898
    sliding-window-rate-limit:123:30:90
    sliding-window-rate-limit:123:30:89
    sliding-window-rate-limit:123:30:88

...none of which exist, so the values are assumed 0 and the request goes
through. It then sets the following keys::

    sliding-window-rate-limit:123:30:90 += 1
    sliding-window-rate-limit:123:3:900 += 1

Another request for the same org happens at time `902`.

* The keys starting with ``:123:30:`` sum up to 1, so the 30-second limit of 100 is not exceeded.
* The keys starting with ``:123:3:`` sum up to 1, so the 3-second limit of 10 is not exceeded.

This request is granted the minimum allowed from the two requested, in this case that is 9 (3-second limit of 10 - 1 used)

Because no quota is exceeded, the request is granted. If one quota summed up to
100 or 10, respectively, the request would be rejected.

When using the quotas, the keys change as follows::

    sliding-window-rate-limit:123:3:900 = 1
    sliding-window-rate-limit:123:3:902 = 1
    sliding-window-rate-limit:123:30:90 = 2

"""

from collections import defaultdict
from dataclasses import dataclass
from time import time
from typing import Iterator, MutableMapping, Optional, Sequence, Tuple, Union

from sentry_redis_tools.clients import RedisCluster, StrictRedis


class InvalidConfiguration(Exception):
    pass


@dataclass(frozen=True)
class Quota:
    # The number of seconds to apply the limit to.
    window_seconds: int

    # A number between 1 and `window_seconds`. Since `window_seconds` is a
    # sliding window, configure what the granularity of that window is.
    #
    # If this is equal to `window_seconds`, the quota resets to 0 every
    # `window_seconds`.  If this is a very small number, the window slides
    # "more smoothly" at the expense of having much more redis keys.
    #
    # The number of redis keys required to enforce a quota is `window_seconds /
    # granularity_seconds`.
    granularity_seconds: int

    #: How many units are allowed within the given window.
    limit: int

    # Override the prefix given by RequestedQuota such that one can implement
    # global limits + per-organization limits. The GrantedQuota will still only
    # contain the prefix of the RequestedQuota
    prefix_override: Optional[str] = None

    def __post__init__(self) -> None:
        assert self.window_seconds % self.granularity_seconds == 0

    def iter_window(self, request_timestamp: int) -> Iterator[int]:
        """
        Iterate over the quota's window, yielding values representing each
        (absolute) granule.

        This function is used to calculate keys for storing the number of
        requests made in each granule.

        The iteration is done in reverse-order (newest timestamp to oldest),
        starting with the key to which a currently-processed request should be
        added. That request's timestamp is `request_timestamp`.

        * `request_timestamp / self.granularity_seconds - 1`
        * `request_timestamp / self.granularity_seconds - 2`
        * `request_timestamp / self.granularity_seconds - 3`
        * ...
        """

        value = request_timestamp // self.granularity_seconds

        for granule_i in range(self.window_seconds // self.granularity_seconds):
            value -= 1
            assert value >= 0, value
            yield value


@dataclass(frozen=True)
class RequestedQuota:
    # A string that all redis state is prefixed with. For example
    # `sentry-string-indexer:123` where 123 is an organization id.
    #
    # Note: You cannot control the redis sharding this way, so curly braces are
    # forbidden.
    prefix: str

    # How much of each quota's limit is requested
    requested: int

    # Which quotas to check against. The requested amount must "fit" into all
    # quotas.
    quotas: Sequence[Quota]


@dataclass(frozen=True)
class GrantedQuota:
    # The prefix from RequestedQuota
    prefix: str

    # How much of RequestedQuota.requested can actually be used.
    granted: int

    # If RequestedQuota.requested > GrantedQuota.granted, this contains the
    # quotas that were reached.
    reached_quotas: Sequence[Quota]


Timestamp = int


class RedisSlidingWindowRateLimiter:
    """
    The main class to instantiate for the rate limiter.

    Example usage::

        quotas = [
            Quota(
                window_seconds=10,
                granularity_seconds=1,
                limit=10,
            )
        ]

        for timestamp in range(10):
            resp = limiter.check_and_use_quotas(
                [RequestedQuota(prefix="foo", requested=1, quotas=quotas)],
                timestamp=TIMESTAMP_OFFSET + timestamp,
            )
            assert resp == [GrantedQuota(prefix="foo", granted=1, reached_quotas=[])]

        resp = limiter.check_and_use_quotas(
            [RequestedQuota(prefix="foo", requested=1, quotas=quotas)],
            timestamp=TIMESTAMP_OFFSET + 9,
        )
        assert resp == [GrantedQuota(prefix="foo", granted=0, reached_quotas=quotas)]
    """

    def __init__(self, client: Union[StrictRedis, RedisCluster]) -> None:
        self.client = client

    def _build_redis_key_raw(
        self, prefix: str, window: int, granularity: int, granule: int
    ) -> str:
        if "{" in prefix or "}" in prefix:
            # The rate limiter currently does not allow you to control the
            # Redis sharding key through the prefix`. This is currently an
            # arbitrary limitation, but the reason for this is that one day we
            # may want to rewrite the internals to run inside of a Lua script
            # to allow for (partially) atomic check-and-use of rate limits (or
            # do that for performance reasons), in which case the rate limiter
            # would have to take control of sharding itself.
            raise ValueError("Explicit sharding not allowed in RequestedQuota.prefix")

        return f"sliding-window-rate-limit:{prefix}:{window}:{granularity}:{granule}"

    def _build_redis_key(
        self, request: RequestedQuota, quota: Quota, granule: int
    ) -> str:
        return self._build_redis_key_raw(
            prefix=quota.prefix_override or request.prefix,
            window=quota.window_seconds,
            granularity=quota.granularity_seconds,
            granule=granule,
        )

    def check_within_quotas(
        self, requests: Sequence[RequestedQuota], timestamp: Optional[Timestamp] = None
    ) -> Tuple[Timestamp, Sequence[GrantedQuota]]:
        """
        Check whether the given requests are within quota, but do not actually consume quota.

        Consuming the quota is done via :py:meth:`use_quotas`, and should be
        done after the protected resource has actually been consumed.
        """
        if timestamp is None:
            timestamp = int(time())
        else:
            timestamp = int(timestamp)

        keys_to_fetch = set()
        for request in requests:
            # We could potentially run this check inside of __post__init__ of
            # RequestedQuota, but the list is actually mutable after
            # construction.
            assert request.quotas

            for quota in request.quotas:
                for granule in quota.iter_window(timestamp):
                    keys_to_fetch.add(
                        self._build_redis_key(
                            request=request, quota=quota, granule=granule
                        )
                    )

        # Stabilize the iteration order of `keys_to_fetch` by converting it
        # into a list, because the next line will iterate over keys_to_fetch
        # twice.
        #
        # While CPython 3.8 does not change the iteration order of a set as
        # long as it is not being modified
        # (https://stackoverflow.com/a/3812600/1544347), there are no formal
        # guarantees about it.
        ordered_keys_to_fetch = list(keys_to_fetch)
        p = self.client.pipeline()
        for k in ordered_keys_to_fetch:
            p.get(k)
        redis_results = dict(zip(ordered_keys_to_fetch, p.execute()))

        results = []

        # for "global quotas" (=quotas using prefix_override, which may be
        # present in multiple requests), we keep a cache of how much quota we
        # have used within this function call
        #
        # the mapping is id(Quota) -> <already granted quota>
        #
        # this prevents us from seriously overcommitting on the global quota,
        # just because each request happens to fit into it
        quota_used_cache: MutableMapping[int, int] = defaultdict(int)

        for request in requests:
            # We start out with assuming the entire request can be granted in
            # its entirety.
            granted_quota = request.requested
            reached_quotas = []

            # A request succeeds (partially) if it fits (partially) into all
            # quotas. For each quota, we calculate how much quota has been used
            # up, and trim the granted_quota by the remaining quota.
            #
            # We need to explicitly handle the possibility that quotas have
            # been overused, in those cases we want to truncate resulting
            # negative "grants" to zero.
            for quota in request.quotas:
                used_quota = (
                    sum(
                        int(
                            redis_results.get(
                                self._build_redis_key(
                                    request=request, quota=quota, granule=granule
                                )
                            )
                            or 0
                        )
                        for granule in quota.iter_window(timestamp)
                    )
                    + quota_used_cache[id(quota)]
                )

                remaining_quota = max(0, quota.limit - used_quota)

                if remaining_quota < granted_quota:
                    granted_quota = remaining_quota
                    reached_quotas.append(quota)

            for quota in request.quotas:
                if quota.prefix_override:
                    quota_used_cache[id(quota)] += granted_quota

            results.append(
                GrantedQuota(
                    prefix=request.prefix,
                    granted=granted_quota,
                    reached_quotas=reached_quotas,
                )
            )

        return timestamp, results

    def use_quotas(
        self,
        requests: Sequence[RequestedQuota],
        grants: Sequence[GrantedQuota],
        timestamp: Timestamp,
    ) -> None:
        """
        Actually use up quotas after checking them using
        :py:meth:`check_within_quotas` and consuming the underlying resource.

        If you don't care for this two-step dance, use :py:meth:`check_and_use_quotas`.
        """
        assert len(requests) == len(grants)

        keys_to_incr: MutableMapping[str, int] = {}
        keys_ttl: MutableMapping[str, int] = {}

        for request, grant in zip(requests, grants):
            assert request.prefix == grant.prefix

            for quota in request.quotas:
                # Only incr most recent granule
                granule = next(quota.iter_window(timestamp))
                key = self._build_redis_key(
                    request=request, quota=quota, granule=granule
                )
                keys_to_incr.setdefault(key, 0)
                keys_to_incr[key] += grant.granted
                keys_ttl[key] = quota.window_seconds

        with self.client.pipeline(transaction=False) as pipeline:
            for key, value in keys_to_incr.items():
                pipeline.incrby(key, value)
                # Expire the key in `window_seconds`. Since the key has been
                # recently incremented we know it represents a current
                # timestamp. We could use expireat here, but in tests we use
                # timestamps starting from 0 for convenience.
                pipeline.expire(key, keys_ttl[key])

            pipeline.execute()

    def check_and_use_quotas(
        self, requests: Sequence[RequestedQuota], timestamp: Optional[Timestamp] = None
    ) -> Sequence[GrantedQuota]:
        """
        Check the quota requests in Redis and consume the quota in one go. See
        `check_within_quotas` for parameters.
        """
        timestamp, grants = self.check_within_quotas(requests, timestamp)
        self.use_quotas(requests, grants, timestamp)
        return grants



import time
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import (
    Collection,
    Iterator,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from sentry_redis_tools.metrics import Metrics
from sentry_redis_tools.clients import BlasterClient, RedisCluster, StrictRedis

Hash = int
Timestamp = int


class Quota(NamedTuple):
    # The number of seconds to apply the limit to.
    window_seconds: int

    # A number between 1 and `window_seconds`. Since `window_seconds` is a
    # sliding window, configure what the granularity of that window is.
    #
    # If this is equal to `window_seconds`, the quota resets to 0 every
    # `window_seconds`.  If this is a very small number, the window slides
    # "more smoothly" at the expense of having much more redis keys.
    #
    # The number of redis keys required to enforce a quota is `window_seconds /
    # granularity_seconds`.
    granularity_seconds: int

    #: How many units are allowed within the given window.
    limit: int

    def __post__init__(self) -> None:
        assert self.window_seconds % self.granularity_seconds == 0

    def iter_window(self, request_timestamp: int) -> Iterator[int]:
        """
        Iterate over the quota's window, yielding values representing each
        (absolute) granule.

        This function is used to calculate keys for storing the number of
        requests made in each granule.

        The iteration is done in reverse-order (newest timestamp to oldest),
        starting with the key to which a currently-processed request should be
        added. That request's timestamp is `request_timestamp`.

        * `request_timestamp / self.granularity_seconds - 1`
        * `request_timestamp / self.granularity_seconds - 2`
        * `request_timestamp / self.granularity_seconds - 3`
        * ...
        """
        value = request_timestamp // self.granularity_seconds

        for granule_i in range(self.window_seconds // self.granularity_seconds):
            value -= 1
            assert value >= 0, value
            yield value


class RequestedQuota(NamedTuple):
    # A string that all redis state is prefixed with. For example
    # `sentry-string-indexer` where 123 is an organization id.
    prefix: str

    # A unit is an abstract term for the object type we want to limit the
    # cardinality of.
    #
    # For example, if you want to limit the cardinality of timeseries in a
    # metrics service, this would be a set of hashes composed from `(org_id,
    # metric_name, tags)`.
    #
    # ...though you can probably omit org_id if it is already in the prefix.
    unit_hashes: Collection[Hash]

    # Which quota to check against. The number of not-yet-seen hashes must
    # "fit" into `Quota.limit`.
    quota: Quota


class GrantedQuota(NamedTuple):
    request: RequestedQuota

    # The subset of hashes provided by the user `self.request` that were
    # accepted by the limiter.
    granted_unit_hashes: Collection[Hash]

    # If len(granted_unit_hashes) < len(RequestedQuota.unit_hashes), this
    # contains the quotas that were reached.
    reached_quota: Optional[Quota]


class CardinalityLimiter(ABC):
    """
    A kind of limiter that limits set cardinality instead of a rate/count.

    The high-level concepts are very similar to `sentry.ratelimits.sliding_windows`.

    Instead of passing in numbers and getting back smaller numbers, however, the
    user passes in a set and gets back a smaller set. Set elements that have
    already been observed in any quota's window are "for free" and will not
    count towards any quota.

    The implementation hasn't been finalized yet, but we expect that under the hood
    this cardinality limiter will be more expensive to operate than a simple rate
    limiter, as it needs to keep track of already-seen set elements. The memory
    usage in Redis will most likely be proportional to the set size.

    This kind of limiter does not support prefix overrides, which practically means
    that there can only be a per-org or a global limit, not both at once.
    """

    @abstractmethod
    def check_within_quotas(
        self, requests: Sequence[RequestedQuota], timestamp: Optional[Timestamp] = None
    ) -> Tuple[Timestamp, Sequence[GrantedQuota]]:
        """
        Given a set of quotas requests and limits, compute how much quota could
        be consumed.

        :param requests: The requests to return "grants" for.
        :param timestamp: The timestamp of the incoming request. Defaults to
            the current timestamp.

            Providing a too old timestamp here _can_ effectively disable rate
            limits, as the older request counts may no longer be stored.
            However, consistently providing old timestamps here will work
            correctly.
        """

        raise NotImplementedError()

    @abstractmethod
    def use_quotas(
        self,
        grants: Sequence[GrantedQuota],
        timestamp: Timestamp,
    ) -> None:
        """
        Given a set of requests and the corresponding return values from
        `check_within_quotas`, consume the quotas.

        :param timestamp: The request timestamp that has previously been passed
            to `check_within_quotas`.
        :param grants: The return value of `check_within_quotas` which
            indicates how much quota should actually be consumed.
        """

        # TODO: we may want to revisit this abstraction once we move the call
        # to `use_quotas` in the string indexer to a separate thread.
        # particularly we are considering to write more recent time windows
        # synchronously while writing older ones on a separate thread.
        raise NotImplementedError()


class RedisCardinalityLimiter(CardinalityLimiter):
    """
    The Redis cardinality limiter stores a key per unit hash, and adds the unit
    hash to a set key prefixed with `RequestedQuota.prefix`. The memory usage
    grows with the dimensionality of `prefix` and the configured quota limit
    per prefix (i.e. the maximum cardinality of
    `GrantedQuota.granted_unit_hashes`).

    Many design decisions for this cardinality limiter were copied from the
    sliding_windows rate limiter. You will find the next couple of paragraphs
    in its documentation as well.

    Why is checking quotas and using quotas two separate implementations?
    Isn't that a time-of-check-time-of-use bug, and allows me to over-spend
    quota when requests are happening concurrently?

    1) It's desirable to first check quotas, then do a potentially fallible
       operation, then consume quotas. This rate limiter is primarily going to
       be used inside of the metrics string indexer to rate-limit database
       writes. What we want to do there is: read DB, check rate limits,
       write to DB, use rate limits.

       If we did it any other way (the obvious one being to read DB,
       check-and-use rate limits, write DB), crashes while writing to the
       database can over-consume quotas. This is not a big problem if those
       crashes are flukes, and especially not a problem if the crashes are
       a result of an overloaded DB.

       It is however a problem in case the consumer is crash-looping, or
       crashing (quickly) for 100% of requests (e.g. due to schema
       mismatches between code and DB that somehow don't surface during the
       DB read). In that case the quotas would be consumed immediately and
       incident recovery would require us to reset all quotas manually (or
       disable rate limiting via some killswitch)

    3) The redis backend (really the only backend we care about) already
       has some consistency problems.

       a) Redis only provides strong consistency and ability to
          check-and-increment counters when all involved keys hit the same
          Redis node. That means that a quota with prefix="org_id:123" can
          only run on a single redis node. It also means that a global
          quota (`prefix="global"`) would have to run on a single
          (unchangeable) redis node to be strongly consistent. That's
          however a problem for scalability.

          There's no obvious way to make global quotas consistent with
          per-org quotas this way, so right now it means that requests
          containing multiple quotas with different `prefixes` cannot be
          checked-and-incremented atomically even if we were to change the
          rate-limiter's interface.

       b) This is easily fixable, but because of the above, we
          currently don't control Redis sharding at all, meaning that even
          keys within a single quota's window will hit different Redis
          node. This also helps further distribute the load internally.

          Since we have given up on atomic check-and-increments in general
          anyway, there's no reason to explicitly control sharding.
    """

    def __init__(
        self,
        client: Union[BlasterClient, StrictRedis, RedisCluster],
        num_shards: int = 3,
        num_physical_shards: int = 3,
        metric_tags: Optional[Mapping[str, str]] = None,
        metrics_backend: Optional[Metrics] = None,
    ) -> None:
        """
        :param cluster: Name of the redis cluster to use, to be configured with
            the `redis.clusters` Sentry option (like any other redis cluster in
            Sentry).
        :param cluster_num_shards: The number of logical shards to have. This
            controls the average set size in Redis.
        :param cluster_num_physical_shards: The number of actual shards to
            store. Controls how many keys of type "unordered set" there are in
            Redis. The ratio `cluster_num_physical_shards / cluster_num_shards`
            is a sampling rate, the lower it is, the less precise accounting
            will be.
        """
        self.backend: RedisBackend = (
            RedisClusterBackend(client)
            if not isinstance(client, BlasterClient)
            else RedisBlasterBackend(client)
        )

        assert 0 < num_physical_shards <= num_shards
        self.num_shards = num_shards
        self.num_physical_shards = num_physical_shards
        self.metric_tags = metric_tags or {}
        self.metrics_backend = metrics_backend
        super().__init__()

    @staticmethod
    def _get_timeseries_key(request: RequestedQuota, hash: Hash) -> str:
        return f"cardinality:timeseries:{request.prefix}-{hash}"

    def _get_read_sets_keys(
        self, request: RequestedQuota, timestamp: Timestamp
    ) -> Sequence[str]:
        oldest_time_bucket = list(request.quota.iter_window(timestamp))[-1]
        return [
            f"cardinality:sets:{request.prefix}-{shard}-{oldest_time_bucket}"
            for shard in range(self.num_physical_shards)
        ]

    def _get_write_sets_keys(
        self, request: RequestedQuota, timestamp: Timestamp, hash: Hash
    ) -> Sequence[str]:
        shard = hash % self.num_shards
        if shard < self.num_physical_shards:
            return [
                f"cardinality:sets:{request.prefix}-{shard}-{time_bucket}"
                for time_bucket in request.quota.iter_window(timestamp)
            ]
        else:
            return []

    def _get_set_cardinality_sample_factor(self) -> float:
        return self.num_shards / self.num_physical_shards

    def check_within_quotas(
        self, requests: Sequence[RequestedQuota], timestamp: Optional[Timestamp] = None
    ) -> Tuple[Timestamp, Sequence[GrantedQuota]]:
        if timestamp is None:
            timestamp = int(time.time())
        else:
            timestamp = int(timestamp)

        unit_keys_to_get: List[str] = []
        set_keys_to_count: List[str] = []

        for request in requests:
            for hash in request.unit_hashes:
                unit_keys_to_get.append(self._get_timeseries_key(request, hash))

            set_keys_to_count.extend(self._get_read_sets_keys(request, timestamp))

        if not unit_keys_to_get and not set_keys_to_count:
            # If there are no keys to fetch (i.e. there are no quotas to
            # enforce), we can save the redis call entirely and just grant all
            # quotas immediately.
            return timestamp, [
                GrantedQuota(
                    request=request,
                    granted_unit_hashes=request.unit_hashes,
                    reached_quota=None,
                )
                for request in requests
            ]

        unit_keys, set_counts = self.backend.run_check_within_quotas(
            unit_keys_to_get, set_keys_to_count
        )

        grants = []
        cardinality_sample_factor = self._get_set_cardinality_sample_factor()
        for request in requests:
            granted_hashes = []

            set_count = sum(
                set_counts[k] for k in self._get_read_sets_keys(request, timestamp)
            )

            if self.metrics_backend:
                self.metrics_backend.timing(
                    "ratelimits.cardinality.set_size",
                    value=set_count,
                    tags=self.metric_tags,
                )

            remaining_limit = max(
                0,
                request.quota.limit - cardinality_sample_factor * set_count,
            )

            remaining_limit_running = remaining_limit
            reached_quota = None

            # for each hash in the request, check if:
            # 1. the hash is in `unit_keys`. If so, it has already been seen in
            #    this timewindow, and ingesting additional copies of it comes
            #    at no cost (= ingesting multiple metric buckets of the same
            #    timeseries only counts once against quota)
            #
            # 2. we still have budget/"remaining_limit". In that case,
            #    accept/admit the hash as well and reduce the remaining quota.
            #
            # 3. we have observed a totally new hash. in that case set
            #    `reached_quotas` for reporting purposes, but don't add the
            #    hash to `granted_hashes` (which is our return value)
            for hash in request.unit_hashes:
                if unit_keys[self._get_timeseries_key(request, hash)]:
                    granted_hashes.append(hash)
                elif remaining_limit_running > 0:
                    granted_hashes.append(hash)
                    remaining_limit_running -= 1
                else:
                    reached_quota = request.quota

            grants.append(
                GrantedQuota(
                    request=request,
                    granted_unit_hashes=granted_hashes,
                    reached_quota=reached_quota,
                )
            )

        return timestamp, grants

    def use_quotas(
        self,
        grants: Sequence[GrantedQuota],
        timestamp: Timestamp,
    ) -> None:
        unit_keys_to_set = {}
        set_keys_to_add = defaultdict(set)
        set_keys_ttl = {}

        for grant in grants:
            key_ttl = grant.request.quota.window_seconds

            for hash in grant.granted_unit_hashes:
                unit_key = self._get_timeseries_key(grant.request, hash)
                unit_keys_to_set[unit_key] = key_ttl

                for set_key in self._get_write_sets_keys(
                    grant.request, timestamp, hash
                ):
                    set_keys_ttl[set_key] = key_ttl
                    set_keys_to_add[set_key].add(hash)

        if not set_keys_to_add and not unit_keys_to_set:
            # If there are no keys to mutate (i.e. there are no quotas to
            # enforce), we can save the redis call entirely.
            return

        self.backend.run_use_quotas(unit_keys_to_set, set_keys_to_add, set_keys_ttl)


class RedisBackend(ABC):
    @abstractmethod
    def run_check_within_quotas(
        self, unit_keys_to_get: Sequence[str], set_keys_to_count: Sequence[str]
    ) -> Tuple[Mapping[str, int], Mapping[str, int]]:
        ...

    @abstractmethod
    def run_use_quotas(
        self,
        unit_keys_to_set: Mapping[str, int],
        set_keys_to_add: Mapping[str, Collection[int]],
        set_keys_ttl: Mapping[str, int],
    ) -> None:
        ...


class RedisClusterBackend(RedisBackend):
    def __init__(self, client: RedisCluster) -> None:
        self.client = client

    def run_check_within_quotas(
        self, unit_keys_to_get: Sequence[str], set_keys_to_count: Sequence[str]
    ) -> Tuple[Mapping[str, int], Mapping[str, int]]:
        with self.client.pipeline(transaction=False) as pipeline:
            for key in unit_keys_to_get:
                pipeline.get(key)
            # O(self.cluster_shard_factor * len(requests)), assuming there's
            # only one per-org quota
            for key in set_keys_to_count:
                pipeline.scard(key)

            results = iter(pipeline.execute())

            unit_keys_results = []
            for _ in range(len(unit_keys_to_get)):
                unit_keys_results.append(next(results))

            unit_keys = dict(zip(unit_keys_to_get, unit_keys_results))
            set_counts = dict(zip(set_keys_to_count, results))

        return unit_keys, set_counts

    def run_use_quotas(
        self,
        unit_keys_to_set: Mapping[str, int],
        set_keys_to_add: Mapping[str, Collection[int]],
        set_keys_ttl: Mapping[str, int],
    ) -> None:
        with self.client.pipeline(transaction=False) as pipeline:
            for key, ttl in unit_keys_to_set.items():
                pipeline.setex(key, ttl, 1)

            for key, items in set_keys_to_add.items():
                items_list = list(items)
                while items_list:
                    # SADD can take multiple arguments, but if you provide too
                    # many you end up with very long-running redis commands.
                    pipeline.sadd(key, *items_list[:200])
                    items_list = items_list[200:]

                pipeline.expire(key, set_keys_ttl[key])

            pipeline.execute()


class RedisBlasterBackend(RedisBackend):
    def __init__(self, client: BlasterClient) -> None:
        self.client = client

    def run_check_within_quotas(
        self, unit_keys_to_get: Sequence[str], set_keys_to_count: Sequence[str]
    ) -> Tuple[Mapping[str, int], Mapping[str, int]]:
        with self.client.map() as client:

            mget_result = client.mget(unit_keys_to_get)

            scard_results = [client.scard(key) for key in set_keys_to_count]

        unit_keys = dict(zip(unit_keys_to_get, mget_result.value))
        set_counts = dict(zip(set_keys_to_count, (p.value for p in scard_results)))

        return unit_keys, set_counts

    def run_use_quotas(
        self,
        unit_keys_to_set: Mapping[str, int],
        set_keys_to_add: Mapping[str, Collection[int]],
        set_keys_ttl: Mapping[str, int],
    ) -> None:
        with self.client.map() as pipeline:
            for key, ttl in unit_keys_to_set.items():
                pipeline.setex(key, ttl, 1)

            for key, items in set_keys_to_add.items():
                items_list = list(items)
                while items_list:
                    # SADD can take multiple arguments, but if you provide too
                    # many you end up with very long-running redis commands.
                    pipeline.sadd(key, *items_list[:200])
                    items_list = items_list[200:]

                pipeline.expire(key, set_keys_ttl[key])



from abc import abstractmethod
from typing import Mapping, Optional, Protocol, Union, runtime_checkable

Tags = Mapping[str, str]


@runtime_checkable
class Metrics(Protocol):
    """
    An abstract class that defines the interface for metrics backends.
    """

    @abstractmethod
    def increment(
        self, name: str, value: Union[int, float] = 1, tags: Optional[Tags] = None
    ) -> None:
        """
        Increments a counter metric by a given value.
        """
        raise NotImplementedError

    @abstractmethod
    def gauge(
        self, name: str, value: Union[int, float], tags: Optional[Tags] = None
    ) -> None:
        """
        Sets a gauge metric to the given value.
        """
        raise NotImplementedError

    @abstractmethod
    def timing(
        self, name: str, value: Union[int, float], tags: Optional[Tags] = None
    ) -> None:
        """
        Records a timing metric.
        """
        raise NotImplementedError


class DummyMetricsBackend(Metrics):
    """
    Default metrics backend that does not record anything.
    """

    def increment(
        self, name: str, value: Union[int, float] = 1, tags: Optional[Tags] = None
    ) -> None:
        pass

    def gauge(
        self, name: str, value: Union[int, float], tags: Optional[Tags] = None
    ) -> None:
        pass

    def timing(
        self, name: str, value: Union[int, float], tags: Optional[Tags] = None
    ) -> None:
        pass



from typing import NoReturn

try:
    from redis import RedisCluster
except ImportError:
    from rediscluster import RedisCluster

try:
    from rb import Cluster as BlasterClient
except ImportError:
    BlasterClient = NoReturn

from redis import StrictRedis

__all__ = ["BlasterClient", "RedisCluster", "StrictRedis"]



from typing import Any
from sentry_redis_tools.clients import RedisCluster

from redis.exceptions import BusyLoadingError, ConnectionError

try:
    from redis.exceptions import ClusterError
except ImportError:
    from rediscluster.exceptions import ClusterError

__all__ = ["ClusterError", "RetryingRedisCluster"]


class RetryingRedisCluster(RedisCluster):  # type: ignore
    """
    Execute a command with cluster reinitialization retry logic.

    Should a cluster respond with a ConnectionError or BusyLoadingError the
    cluster nodes list will be reinitialized and the command will be executed
    again with the most up to date view of the world.
    """

    def execute_command(self, *args: Any, **kwargs: Any) -> Any:
        try:
            return super(self.__class__, self).execute_command(*args, **kwargs)
        except (
            ConnectionError,
            BusyLoadingError,
            ClusterError,
            KeyError,  # see: https://github.com/Grokzen/redis-py-cluster/issues/287
        ):
            # Codepath for redis-py-cluster
            if hasattr(self, "connection_pool"):
                self.connection_pool.nodes.reset()

            # Codepath for redis.cluster
            #
            # the code in the RedisCluster __init__ idiotically sets
            # self.nodes_manager = None
            # self.nodes_manager = NodesManager(...)
            if hasattr(self, "nodes_manager"):
                self.nodes_manager.reset()

            return super(self.__class__, self).execute_command(*args, **kwargs)



from setuptools import find_packages, setup

from typing import Sequence


setup(
    name="sentry-redis-tools",
    version="0.1.7",
    author="Sentry",
    author_email="oss@sentry.io",
    license="Apache-2.0",
    url="https://github.com/getsentry/sentry-redis-tools",
    description="Common utilities related to how Sentry uses Redis",
    zip_safe=False,
    install_requires=['redis>=3.0'],
    packages=find_packages(exclude=("tests", "tests.*")),
    package_data={"sentry_redis_tools": ["py.typed"]},
    include_package_data=True,
    options={"bdist_wheel": {"universal": "1"}},
)



from __future__ import annotations

import random
import time
from typing import Any, Callable, Optional

from sentry_redis_tools.clients import StrictRedis
from redis.exceptions import (
    ConnectionError,
    ReadOnlyError,
    TimeoutError,
)


def _sentry_wrap_with_retry(
    get_wrapped_fn: Callable[[], Any], client_self: Optional[FailoverRedis] = None
) -> Any:
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        nonlocal client_self
        slf = client_self
        if slf is None:
            slf = args[0]
        assert slf is not None

        retries = 0
        while True:
            try:
                return get_wrapped_fn()(*args, **kwargs)
            except (
                # Caught during the inital phase of failver when writes are
                # paused on primary
                ReadOnlyError,
                # When the connection to primary is dropped and the one to the
                # replica is not ready yet.
                # ConnectionError with the errno ETIMEDOUT = 110
                ConnectionError,
                # When the client is initiated with socket_timeout or
                # socket_connect_timeout, during the reconnect it throws
                # redis.exceptions.TimeoutError instead of ConnectionError
                TimeoutError,
            ):
                if retries >= slf._retries:
                    raise
                time.sleep(
                    min(
                        slf._backoff_max,
                        (slf._backoff_min * (slf._backoff_multiplier**retries))
                        * (1 + random.random()),
                    )
                )
                retries += 1

    return wrapper


class FailoverRedis(StrictRedis):  # type: ignore
    """
    Single host redis client implementation with retry logic intended to
    survive failover events. Retry logic uses capped exponential backoff with
    jitter.

    https://redis.io/commands/failover

    Failover sequence:

    1. The primary will internally start a CLIENT PAUSE WRITE, which will pause
    incoming writes and prevent the accumulation of new data in the replication
    stream. From this point all writes to the primary instance fails with
    ReadOnlyError.

    2. The primary will monitor its replicas, waiting for a replica to indicate
    that it has fully consumed the replication stream. If the primary has
    multiple replicas, it will only wait for the first replica to catch up.

    3. The primary will then demote itself to a replica. This is done to
    prevent any dual master scenarios.

    4. The previous primary will send a special PSYNC request to the target
    replica, PSYNC FAILOVER, instructing the target replica to become a
    primary.

    5. Once the previous primary receives acknowledgement the PSYNC FAILOVER
    was accepted it will unpause its clients.

    In addition, the Memorystore for Redis, which is the main target of this implementation states:

    When the primary node fails over to the replica, existing connections to
    the primary endpoint of the instance are dropped. The instance is
    unavailable for a few seconds while the new primary reconnects. On
    reconnect, your application is automatically redirected to the new primary
    node using the same connection string or IP address. You do not need to
    update your application after a failover.

    https://cloud.google.com/memorystore/docs/redis/high-availability#how_a_failover_affects_your_application

    """

    def __init__(
        self,
        *args: Any,
        _retries: int = 10,
        _backoff_min: float = 0.2,
        _backoff_max: float = 5,
        _backoff_multiplier: float = 2,
        **kwargs: Any,
    ):
        if _retries < 0:
            raise ValueError(
                f"Number of retries must non negative integer: _retries={_retries}"
            )
        self._retries = _retries

        if _backoff_min < 0.0:
            raise ValueError(
                f"Minimal backoff must be non negative number: _backoff_min={_backoff_min}"
            )
        self._backoff_min = _backoff_min

        if _backoff_max < _backoff_min:
            raise ValueError(
                f"Maximal backoff must be at least equal to the minimal ({_backoff_min}): _backoff_max={_backoff_max}"
            )
        self._backoff_max = _backoff_max

        if _backoff_multiplier <= 0:
            raise ValueError(
                f"Backoff multiplier must be positive number: _backoff_multiplier={_backoff_multiplier}"
            )
        self._backoff_multiplier = _backoff_multiplier
        super().__init__(*args, **kwargs)

    execute_command = _sentry_wrap_with_retry(lambda: StrictRedis.execute_command)

    def pipeline(self, *args: Any, **kwargs: Any) -> Any:
        rv = StrictRedis.pipeline(self, *args, **kwargs)
        old_execute = rv.execute
        rv.execute = _sentry_wrap_with_retry(lambda: old_execute, client_self=self)
        return rv
